"""
The parent class to simulate raw data for SIM.
Some calculations are adapted by the work by Mark Neil @Imperial College London
"""
__author__ = "Meizhu Liang @Imperial College London"

import numpy as np
import opt_einsum as oe
import time

try:
    import cupy as cp

    print('cupy imported')
    import_cp = True
except:
    import_cp = False

try:
    import torch

    print('torch imported')
    import_torch = True
    if torch.has_cuda:
        torch_GPU = True
    else:
        torch_GPU = False
except:
    import_torch = False
    torch_GPU = False


class Base_simulator:
    points = None
    npoints = 0
    xp = np
    acc = None  # acceleration
    pol = None  # polarisation
    psf_calc = None
    _tdev = None
    N = 512  # points to use in FFT
    pixel_size = 5.5  # camera pixel size
    magnification = 60  # objective magnification
    ill_NA = 0.75  # numerical aperture at illumination beams
    det_NA = 1.1  # numerical aperture at sample
    n = 1.33  # refractive index at sample
    ill_wavelength = 520e-3  # illumination wavelength in um
    det_wavelength = 570e-3  # detection wavelength in um
    zrangeN = 6  # Nyquist distance either side of focus to calculate in microns (depth of the sample)
    # fwhmz = 3.0  # FWHM of light sheet in z
    random_seed = 123
    drift = 0
    spherical = 0  # spherical aberration in um
    sph_abb = 0
    add_error = False

    def initialise(self):
        if self.acc == 3:
            self.xp = cp
        np.random.seed(self.random_seed)
        # self.seed(1234)  # set random number generator seed
        # self.sigmaz = self.fwhmz / 2.355
        self.dx = self.pixel_size / self.magnification  # Sampling in lateral plane at the sample in um
        self.res = self.det_wavelength / (2 * self.det_NA)
        self.dxn = self.res / 2  # 2 * Nyquist frequency in x and y.
        self.Nn = int(np.ceil(self.N * self.dx / self.dxn / 2) * 2)  # Number of points at Nyquist sampling, even number
        self.dxn = self.N * self.dx / self.Nn  # correct spacing
        oversampling = self.res / self.dxn  # factor by which pupil plane oversamples the coherent psf data
        self.k0_ill = 2 * np.pi * self.n / self.ill_wavelength
        self.k0_det = 2 * np.pi * self.n / self.det_wavelength
        self.krmax = self.det_NA * self.k0_det / self.n
        self.dk = oversampling / (self.Nn / 2) * self.krmax  # Pupil plane sampling
        self.kx, self.ky = self.xp.meshgrid(
            self.xp.linspace(-self.dk * self.Nn / 2, self.dk * self.Nn / 2 - self.dk, self.Nn),
            self.xp.linspace(-self.dk * self.Nn / 2, self.dk * self.Nn / 2 - self.dk, self.Nn))
        self.kr = np.sqrt(self.kx ** 2 + self.ky ** 2)  # Raw pupil function, pupil defined over circle of radius self.krmax.
        self.spherical = self.sph_abb * np.sqrt(5) * (6 * ((self.kr / self.krmax) ** 4 - (self.kr / self.krmax) ** 2) + 1)
        # Nyquist sampling in z
        self.dzn = self.det_wavelength / (2 * self.n * (1 - np.cos(np.arcsin(self.det_NA / self.n))))
        self.Nzn = int(2 * np.ceil(self.zrangeN / self.dzn))
        self.dzn = 2 * self.zrangeN / self.Nzn  # step size in axial direction of PSF in um
        if (self.acc == 1) | (self.acc == 2):
            self._tdev = torch.device('cuda' if self.acc == 2 else 'cpu')

        # pre allocate phasetilts arrays
        if (self.acc == 0) or (self.acc == 3):
            if self.psf_calc == 'vector_rigid':
                self.phasetilts = [self.xp.zeros((3, self.Nzn, self.Nn, self.Nn), dtype=self.xp.complex64)
                                   for i in range(self._nsteps)]
            else:
                self.phasetilts = [self.xp.zeros((self.Nzn, self.Nn, self.Nn), dtype=self.xp.complex64)
                                   for i in range(self._nsteps)]
        else:
            if self.psf_calc == 'vector_rigid':
                self.phasetilts = [torch.zeros(3, self.Nzn, self.Nn, self.Nn, dtype=torch.complex64,
                                               device=self._tdev) for i in range(self._nsteps)]
            else:
                self.phasetilts = [torch.zeros(self.Nzn, self.Nn, self.Nn, dtype=torch.complex64,
                                               device=self._tdev) for i in range(self._nsteps)]

    def phase_tilts(self):
        """Generates phase tilts in frequency space"""
        xyrange = self.Nn / 2 * self.dxn
        dkxy = np.pi / xyrange
        dkz = np.pi / self.zrangeN
        if (self.acc == 0) or (self.acc == 3):
            kxy = self.xp.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy, dtype=self.xp.single)
            kz = self.xp.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz, dtype=self.xp.single)
        else:
            kxy = torch.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy,
                               dtype=torch.float32, device=self._tdev)
            kz = torch.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz,
                              dtype=torch.float32, device=self._tdev)

        start_time = time.time()
        itcount = 0
        total_its = self._angleStep * self._phaseStep * self.npoints
        lastProg = 0

        for astep in range(self._angleStep):
            self.jones_vectors(astep)
            for pstep in range(self._phaseStep):
                self.points += self.drift * np.random.standard_normal(3) / 1000
                self.points[:, 0] += self.xdrift / 1000
                if self.zdrift:
                    self.points[:, 2] += self.zdrift / 1000
                istep = pstep + self._phaseStep * astep  # index of the steps
                prog = (100 * itcount) // total_its
                if prog > lastProg + 9:
                    lastProg = prog
                    yield f'Phase tilts calculation: {prog:.1f}% done'
                itcount += 1

                # get illumination from the child class

                if (self.acc == 0) or (self.acc == 3):
                    x = self.xp.array(self.points[:, 0], dtype=self.xp.single)
                    y = self.xp.array(self.points[:, 1], dtype=self.xp.single)
                    z = self.xp.array(self.points[:, 2], dtype=self.xp.single)
                    px = self.xp.exp(1j * kxy[self.xp.newaxis, :] * x[:, self.xp.newaxis])
                    py = self.xp.exp(1j * kxy[self.xp.newaxis, :] * y[:, self.xp.newaxis])
                    pz = self.xp.exp(1j * kz[self.xp.newaxis, :] * z[:, self.xp.newaxis])
                    if self.psf_calc == 'vector_rigid':
                        ill = self.xp.array(self._ill_obj_vec(x, y, z, pstep, astep),
                                            dtype=self.xp.single)
                        oe.contract('im,il,ik,ij->mjkl', ill, px, py, pz, out=self.phasetilts[istep])
                    else:
                        ill = self.xp.array(self._ill_obj(x, y, z, pstep, astep),
                                            dtype=self.xp.single)
                        oe.contract('i,il,ik,ij->jkl', ill, px, py, pz, out=self.phasetilts[istep])
                else:
                    x = torch.tensor(self.points[:, 0], dtype=torch.float32, device=self._tdev)
                    y = torch.tensor(self.points[:, 1], dtype=torch.float32, device=self._tdev)
                    z = torch.tensor(self.points[:, 2], dtype=torch.float32, device=self._tdev)

                    px = torch.exp(1j * kxy[None, :] * x[:, None])
                    py = torch.exp(1j * kxy[None, :] * y[:, None])
                    pz = torch.exp(1j * kz[None, :] * z[:, None])
                    if self.psf_calc == 'vector_rigid':
                        ill = torch.tensor(self._ill_obj_vec(self.points[:, 0], self.points[:, 1], self.points[:, 2], pstep, astep),
                                           dtype=torch.float32, device=self._tdev)
                        oe.contract('im,il,ik,ij->mjkl', ill, px, py, pz, out=self.phasetilts[istep])
                    else:
                        ill = torch.tensor(self._ill_obj(self.points[:, 0], self.points[:, 1], self.points[:, 2], pstep, astep),
                                           dtype=torch.float32, device=self._tdev)
                        oe.contract('i,il,ik,ij->jkl', ill, px, py, pz, out=self.phasetilts[istep])
        self.elapsed_time = time.time() - start_time
        yield f'Phase tilts calculation time:  {self.elapsed_time:3f}s'

    def get_vector_psf(self):
        # use krmax to define the pupil function
        kx = self.kx + 1e-7  # need to add small offset to avoid division by zero
        ky = self.ky + 1e-7
        kr2 = kx ** 2 + ky ** 2  # square kr
        pupil = 1.0 * (kr2 < self.krmax ** 2)  # change of the pupil in the amplitude
        kz = np.sqrt((self.k0_det ** 2 - kr2) + 0j)

        # Calculating psf
        nz = 0
        psf_x = self.xp.zeros((self.Nzn, self.Nn, self.Nn))
        psf_y = self.xp.zeros((self.Nzn, self.Nn, self.Nn))
        psf_z = self.xp.zeros((self.Nzn, self.Nn, self.Nn))

        # calculate intensity of random arrangement of dipoles excited by a given polarisation s
        # p are the vertices of an dodecahedron
        p0 = self.xp.reshape(self.xp.array([0, 1, 0]), (1, 3))
        p1 = self.xp.reshape(self.xp.array([-0.666666, 0., 0.745353,
                                            0.666666, 0., -0.745353,
                                            -0.127322, -0.93417, 0.333332,
                                            -0.127322, 0.93417, 0.333332,
                                            0.745355, -0.577349, -0.333332,
                                            0.745355, 0.577349, -0.333332,
                                            0.333332, -0.577349, 0.745355,
                                            0.333332, 0.577349, 0.745355,
                                            -0.872676, -0.356821, -0.333334,
                                            -0.872676, 0.356821, -0.333334,
                                            0.872676, -0.356821, 0.333334,
                                            0.872676, 0.356821, 0.333334,
                                            1.46634 * 1e-6, 0., -0.999998,
                                            -0.745355, -0.577349, 0.333332,
                                            -0.745355, 0.577349, 0.333332,
                                            -1.46634 * 1e-6, 0., 0.999998,
                                            -0.333332, -0.577349, -0.745355,
                                            -0.333332, 0.577349, -0.745355,
                                            0.127322, -0.93417, -0.333332,
                                            0.127322, 0.93417, -0.333332]), (20, 3))

        # p2 are the vertices of the same dodecahedron in a different orientation
        p2 = self.xp.reshape(self.xp.array([-1.37638, 0., 0.262866,
                                            1.37638, 0., -0.262866,
                                            -0.425325, -1.30902, 0.262866,
                                            -0.425325, 1.30902, 0.262866,
                                            1.11352, -0.809017, 0.262866,
                                            1.11352, 0.809017, 0.262866,
                                            -0.262866, -0.809017, 1.11352,
                                            -0.262866, 0.809017, 1.11352,
                                            -0.688191, -0.5, -1.11352,
                                            -0.688191, 0.5, -1.11352,
                                            0.688191, -0.5, 1.11352,
                                            0.688191, 0.5, 1.11352,
                                            0.850651, 0., -1.11352,
                                            -1.11352, -0.809017, -0.262866,
                                            -1.11352, 0.809017, -0.262866,
                                            -0.850651, 0., 1.11352,
                                            0.262866, -0.809017, -1.11352,
                                            0.262866, 0.809017, -1.11352,
                                            0.425325, -1.30902, -0.262866,
                                            0.425325, 1.30902, -0.262866]), (20, 3))
        # got the flattened array and calculate the square root of the sum of squares
        p2 = p2 / self.xp.linalg.norm(p2[0, :])

        p = p2[p2[:, 2] > 0]  # just select one half of the orientations as their emission is symmetric
        plen = p.shape[0]

        s1 = self.xp.array([1, 0, 0])  # x polarised illumination orientation
        excitation1 = (s1 @ p.T) ** 2
        s2 = self.xp.array([0, 1, 0])  # y polarised illumination orientation
        excitation2 = (s2 @ p.T) ** 2
        s3 = self.xp.array([0, 0, 1])  # z polarised illumination orientation
        excitation3 = (s3 @ p.T) ** 2

        fx1 = self.xp.sqrt(self.k0_det / kz) * \
              (self.k0_det * ky ** 2 + kx ** 2 * kz) / (self.k0_det * kr2) * pupil
        fy1 = self.xp.sqrt(self.k0_det / kz) * \
              kx * ky * (kz - self.k0_det) / (self.k0_det * kr2) * pupil
        fx2 = self.xp.sqrt(self.k0_det / kz) * \
              kx * ky * (kz - self.k0_det) / (self.k0_det * kr2) * pupil
        fy2 = self.xp.sqrt(self.k0_det / kz) * \
              (self.k0_det * kx ** 2 + ky ** 2 * kz) / (self.k0_det * kr2) * pupil
        fx3 = self.xp.sqrt(self.k0_det / kz) * kx / self.k0_det * pupil
        fy3 = self.xp.sqrt(self.k0_det / kz) * ky / self.k0_det * pupil

        for z in np.arange(-self.zrangeN, self.zrangeN, self.dzn):
            pupil_phase = np.exp(1j * (z * kz + self.spherical))  # change of the pupil in the phase
            # x-polarised field at camera for x-oriented dipole
            Exx = self.xp.fft.fftshift(
                self.xp.fft.ifft2(fx1 * pupil_phase))
            # y-polarised field at camera for x-oriented dipole
            Exy = self.xp.fft.fftshift(
                self.xp.fft.ifft2(fy1 * pupil_phase))
            # x-polarised field at camera for y-oriented dipole
            Eyx = self.xp.fft.fftshift(
                self.xp.fft.ifft2(fx2 * pupil_phase))
            # y-polarised field at camera for y-oriented dipole
            Eyy = self.xp.fft.fftshift(
                self.xp.fft.ifft2(fy2 * pupil_phase))
            # x-polarised field at camera for z-oriented dipole
            Ezx = self.xp.fft.fftshift(
                self.xp.fft.ifft2(fx3 * pupil_phase))
            # y-polarised field at camera for z-oriented dipole
            Ezy = self.xp.fft.fftshift(
                self.xp.fft.ifft2(fy3 * pupil_phase))
            intensityx = self.xp.zeros((self.Nn, self.Nn))
            intensityy = self.xp.zeros((self.Nn, self.Nn))
            intensityz = self.xp.zeros((self.Nn, self.Nn))
            for i in np.arange(p.shape[0]):
                intensityx = intensityx + excitation1[i] * (abs(p[i, 0] * Exx + p[i, 1] * Eyx + p[i, 2] * Ezx) ** 2
                                                            + abs(p[i, 0] * Exy + p[i, 1] * Eyy + p[i, 2] * Ezy) ** 2)
                intensityy = intensityy + excitation2[i] * (abs(p[i, 0] * Exx + p[i, 1] * Eyx + p[i, 2] * Ezx) ** 2 +
                                                            abs(p[i, 0] * Exy + p[i, 1] * Eyy + p[i, 2] * Ezy) ** 2)
                intensityz = intensityz + excitation3[i] * (abs(p[i, 0] * Exx + p[i, 1] * Eyx + p[i, 2] * Ezx) ** 2 +
                                                            abs(p[i, 0] * Exy + p[i, 1] * Eyy + p[i, 2] * Ezy) ** 2)

            psf_x[nz, :, :] = intensityx
            psf_y[nz, :, :] = intensityy
            psf_z[nz, :, :] = intensityz
            nz = nz + 1
        psf_x = psf_x * self.Nn ** 2 / self.xp.sum(pupil) / plen
        psf_y = psf_y * self.Nn ** 2 / self.xp.sum(pupil) / plen
        psf_z = psf_z * self.Nn ** 2 / self.xp.sum(pupil) / plen
        return psf_x, psf_y, psf_z

    def get_scalar_psf(self):
        # use krmax to define the pupil function
        nz = 0
        psf = self.xp.zeros((self.Nzn, self.Nn, self.Nn))
        pupil = self.kr < (self.krmax)
        kz = np.sqrt(self.k0_det ** 2 - (self.kr * pupil) ** 2)
        for z in np.arange(-self.zrangeN, self.zrangeN, self.dzn):
            c = np.exp(1j * (z * kz + self.spherical)) * pupil
            psf[nz, :, :] = abs(np.fft.fftshift(np.fft.ifft2(c, norm='ortho'))) ** 2
            # default 'backwards' normalisation: psf[nz, :, :] = abs(np.fft.fftshift(np.fft.ifft2(c))) ** 2
            nz = nz + 1
        # Normalised so power in resampled psf(see later on) is unity in focal plane
        psf = psf / self.xp.sum(pupil ** 2)
        # normalisation factor: Nn ** 2, for use with norm='backwards' ifft:
        # psf = psf * (self.Nn **2 / self.xp.sum(pupil ** 2))
        return psf

    def raw_image_stack_brownian(self):
        # Calculates 3d psf and otf before the image stack
        self.initialise()
        # Calculating psf
        if self.psf_calc == 'vector_rigid':
            psf_x, psf_y, psf_z = self.get_vector_psf()
            norm_val = np.sum(psf_x, axis=(1, 2)).max()
            psf_x /= norm_val
            psf_y /= norm_val
            psf_z /= norm_val
            psf = psf_z
            self.psf_x = psf_x
            self.psf_y = psf_y
            self.psf_z = psf_z
        elif self.psf_calc == 'vector_flexible':
            psf_x, psf_y, psf_z = self.get_vector_psf()
            psf = psf_x + psf_y + psf_z
            norm_val = np.sum(psf, axis=(1, 2)).max()
            psf /= norm_val
        else:
            psf = self.get_scalar_psf()

        yield "psf calculated"
        self.psffile = np.sum(psf, axis=(1, 2))
        # Calculating 3d otf
        self.psf = psf  # 3d psf for display
        if self.acc == 3:
            self.psf = cp.asnumpy(self.psf)
        psf = self.xp.fft.fftshift(psf, axes=0)  # need to set plane zero as in-focus here
        otf = self.xp.fft.fftn(psf)
        if self.psf_calc == 'vector_rigid':
            otf_x = self.xp.fft.fftn(self.xp.fft.fftshift(psf_x, axes=0))  # need to set plane zero as in-focus here
            otf_y = self.xp.fft.fftn(self.xp.fft.fftshift(psf_y, axes=0))  # need to set plane zero as in-focus here
            otf_z = self.xp.fft.fftn(self.xp.fft.fftshift(psf_z, axes=0))  # need to set plane zero as in-focus here
        aotf = abs(self.xp.fft.fftshift(otf))  # absolute otf, also modulus of otf (MTF)
        self.aotffile = np.sum(aotf, axis=(1, 2))
        if self.acc == 3:
            aotf = cp.asnumpy(aotf)
        m = max(aotf.flatten())
        self.aotf = np.log(aotf + aotf.max() / 1000)  # 3d aotf with log
        if self.acc == 3:
            self.aotf = cp.asnumpy(self.aotf)
        yield "3d otf calculated"

        self.points[:, 0] -= self.xdrift * self.tpoints / 2000
        if self.zdrift:
            self.points[:, 2] -= self.zdrift * self.tpoints / 2000
        else:
            self.points[:, 2] -= self.zstep * (self.tpoints / self._nsteps) / 2000
        if (self.acc == 0) | (self.acc == 3):
            img = self.xp.zeros((int(self.tpoints), self.N, self.N), dtype=np.single)
        else:
            img = torch.empty((int(self.tpoints), self.N, self.N), dtype=torch.float, device=self._tdev)

        start_Brownian = time.time()
        tplane = 0
        yield "Starting stack"

        for t in np.arange(0, self.tpoints, self._nsteps):
            for msg in self.phase_tilts():
                yield f'planes at time point = {t} of {self.tpoints}, {msg}'
            if self.zstep:
                self.points[:, 2] += self.zstep / 1000
            for i in range(self._nsteps):
                if (self.acc == 0) | (self.acc == 3):
                    if self.psf_calc == 'vector_rigid':
                        ootf = self.xp.fft.fftshift(otf_x) * self.phasetilts[i][0, :, :, :] + \
                               self.xp.fft.fftshift(otf_y) * self.phasetilts[i][1, :, :, :] + \
                               self.xp.fft.fftshift(otf_z) * self.phasetilts[i][2, :, :, :]
                    else:
                        ootf = self.xp.fft.fftshift(otf) * self.phasetilts[i]
                    img[tplane, :, :] = self.xp.abs(
                        self.xp.fft.ifft2(self.xp.sum(ootf, axis=0), (self.N, self.N)))
                else:
                    if self.psf_calc == 'vector_rigid':
                        ootf = torch.fft.fftshift(torch.as_tensor(otf_x, device=self._tdev)) * \
                               self.phasetilts[i][0, :, :, :] + \
                               torch.fft.fftshift(torch.as_tensor(otf_y, device=self._tdev)) * \
                               self.phasetilts[i][1, :, :, :] + \
                               torch.fft.fftshift(torch.as_tensor(otf_z, device=self._tdev)) * \
                               self.phasetilts[i][2, :, :, :]
                    else:
                        ootf = torch.fft.fftshift(torch.as_tensor(otf, device=self._tdev)) * self.phasetilts[i]
                    img[tplane, :, :] = (torch.abs(
                        torch.fft.ifft2(torch.sum(ootf, axis=0), (self.N, self.N)))).to(torch.float)
                tplane += 1

        # OK to use abs here as signal should be all positive.
        # Abs is required as the result will be complex as the fourier plane cannot be shifted back to zero when oversampling.
        # But should reduction in sampling be allowed here(Nz < Nzn)?

        # stackfilename = f"Raw_img_stack_{self.N}_{self.pol}.tif"
        if self.acc == 0:
            # raw image stack
            # raw image sum along z axis
            self.img_sum_z = np.sum(img, axis=0)
            # raw image sum along x (or y) axis
            self.img_sum_x = np.sum(img, axis=1)
            self.img = img
        elif self.acc == 1:
            self.img_sum_z = (torch.sum(img, axis=0)).numpy()
            self.img_sum_x = (torch.sum(img, axis=1)).numpy()
            self.img = img.numpy()
        elif self.acc == 2:
            self.img_sum_z = (torch.sum(img, axis=0)).detach().cpu().numpy()
            self.img_sum_x = (torch.sum(img, axis=1)).detach().cpu().numpy()
            self.img = img.detach().cpu().numpy()
        elif self.acc == 3:
            self.img_sum_z = cp.asnumpy(cp.sum(img, axis=0))
            self.img_sum_x = cp.asnumpy(cp.sum(img, axis=1))
            self.img = cp.asnumpy(img)

        # # Save generated images
        # tifffile.imwrite(stackfilename, self.img)
        elapsed_Brownian = time.time() - start_Brownian
        yield f'Finished, Phase tilts calculation time:  {elapsed_Brownian:3f}s'

    def illumination_stack(self):
        # Calculates 3d psf and otf before the image stack
        print(f'allocating illumination stack: {(int(self.tpoints), self.N, self.N)}')

        illumination = self.xp.zeros((int(self.tpoints), self.N, self.N))
        xyvals = (self.xp.arange(self.N) - self.N / 2) * self.dx
        xarr, yarr = self.xp.meshgrid(xyvals, xyvals)
        xarr_l, yarr_l = self.xp.array(xarr).flatten(), self.xp.array(yarr).flatten()
        zarr_l = self.xp.zeros_like(xarr_l)
        self.npoints = xarr_l.shape[0]

        start_time = time.time()
        itcount = -1
        for astep in range(self._angleStep):
            self.jones_vectors(astep)
            for pstep in range(self._phaseStep):
                itcount += 1
                ill_1d = self._ill_obj(xarr_l, yarr_l, zarr_l, pstep, astep)
                illumination[itcount, :, :] = self.xp.reshape(ill_1d, (self.N, self.N))

        for i in range(1, int(self.tpoints) // self._nsteps):
            for j in range(self._nsteps):
                illumination[i * self._nsteps + j, :, :] = illumination[j, :, :]

        elapsed_time = time.time() - start_time
        print(f'Illumination calculation time:  {elapsed_time:3f}s')
        if self.acc == 3:
            return cp.asnumpy(illumination)
        else:
            return illumination



"""
Classes to calculate SIM illumination.
"""
__author__ = "Meizhu Liang @Imperial College London"

from .baseSIMulator import Base_simulator
import numpy as np
from numpy import cos, sin


class Illumination(Base_simulator):
    """A class to calculate illumination patterns of multiple beams."""

    def __init__(self):
        super().__init__()
        self._nsteps = int(self._phaseStep * self._angleStep)
        # systematic errors, could be arbitrary
        if self.add_error:
            self.phase_error = np.reshape((2 * np.random.random(int(self._nsteps * self._n_beams)) - 1),
                                          (self._n_beams, self._angleStep, self._phaseStep))
            self.phase_error[0] = 0.0
            self.phase_error[:, :, 0] = 0.0
            # print(f'phase errors:{self.phase_error}')
            self.angle_error = np.reshape((2 * np.random.random(self._angleStep * self._n_beams) - 1),
                                          (self._n_beams, self._angleStep))
            self.angle_error[0] = 0.0
            self.angle_error[:, 0] = 0.0
            # print(f'angle errors:{self.angle_error}')
        else:
            self.phase_error = np.zeros((self._n_beams, self._angleStep, self._phaseStep))
            self.angle_error = np.zeros((self._n_beams, self._angleStep))

    def rotation(self, phi, theta):
        """Calculates the rotation matrix"""
        R = self.xp.array([[cos(phi), -sin(phi), 0], [sin(phi), cos(phi), 0], [0, 0, 1]]) \
            @ self.xp.array([[cos(theta), 0, -sin(theta)], [0, 1, 0], [sin(theta), 0, cos(theta)]]) \
            @ self.xp.array([[cos(phi), sin(phi), 0], [-sin(phi), cos(phi), 0], [0, 0, 1]])
        return R

    def polarised_field(self, phi):
        if self.pol == 'a':
            f_p = self.xp.array([[cos(phi), -sin(phi)], [sin(phi), cos(phi)], [0, 0]]) @ self.xp.array([[0], [1]])
        elif self.pol == 'r':
            f_p = self.xp.array([[cos(phi), -sin(phi)], [sin(phi), cos(phi)], [0, 0]]) @ self.xp.array([[1], [0]])
        elif self.pol == 'c':
            f_p = self.xp.array([[1, 0], [0, 1], [0, 0]]) @ (self.xp.array([[1], [1j]]) / self.xp.sqrt(2))
        elif self.pol == 'h':
            f_p = self.xp.array([[1, 0], [0, 1], [0, 0]]) @ self.xp.array([[1], [0]])
        elif self.pol == 'v':
            f_p = self.xp.array([[1, 0], [0, 1], [0, 0]]) @ self.xp.array([[0], [1]])
        return f_p

    def jones_vectors(self, astep):
        self.theta = np.arcsin(self.ill_NA / self.n)
        self.S = self.xp.zeros((self._n_beams, 3), dtype=self.xp.complex64)
        for i in range(self._n_beams):
            phi_S = i * self._beam_a + astep * 2 * self.xp.pi / self._angleStep
            f_p = self.xp.array(self.polarised_field(phi_S))
            self.S[i, :] = self.xp.transpose(self.rotation(phi_S, self.theta) @ f_p)

    def _ill_obj(self, x, y, z, pstep, astep):
        """Illumination intensity applied on the object"""
        ill = self.xp.sum(self._ill_obj_vec(x, y, z, pstep, astep), axis=1)  # take real part and round to 15 decimals
        return ill

    def _ill_obj_vec(self, x, y, z, pstep, astep):
        """Vectorised illumination intensity applied on the object"""
        p = [0, pstep * 2 * np.pi / self._phaseStep, pstep * (-4) * np.pi / self._phaseStep]
        E = self.xp.zeros((self.npoints, self._n_beams, 3), dtype=self.xp.complex64)  # exponential terms of field
        xyz = self.xp.transpose(self.xp.stack([x, y, z]))
        for i in range(self._n_beams):
            phi_E = i * self._beam_a + astep * 2 * np.pi / self._angleStep + self.angle_error[i, astep]
            e = self.xp.exp(-1j * (xyz @ self.rotation(phi_E, self.theta) @ self.xp.array([0, 0, self.k0_ill]) + p[i] +
                                   self.phase_error[i, astep, pstep]))
            E[:, i, :] = self.xp.transpose(self.xp.array([e, ] * 3))
        F = self.xp.sum(self.S * E, axis=1, dtype=self.xp.complex64)  # field of illumination
        # to calculate intensity: ill = F @ F_conjugate
        ill = (F * self.xp.conjugate(F)).real.round(15)  # take real part and round to 15 decimals
        return ill

class ConIll(Illumination):
    def __init__(self):
        self._phaseStep = 3
        self._angleStep = 3
        self._n_beams = 2
        self._beam_a = 2 * np.pi / self._n_beams  # angle between each two beams
        self._nbands = 3
        super().__init__()


class HexIll(Illumination):
    def __init__(self):
        self._phaseStep = 7
        self._angleStep = 1
        self._n_beams = 3
        self._beam_a = 2 * np.pi / self._n_beams  # angle between each two beams
        self._nbands = 3
        super().__init__()


class RaHexIll(Illumination):
    def __init__(self):
        self._phaseStep = 5
        self._angleStep = 1
        self._n_beams = 3
        self._beam_a = np.pi / 2  # angle between beam1 and beam2, beam2 and beam3
        self._nbands = 3
        super().__init__()

class ConIll3D(Illumination):
    """
    Conventional three-dimensional SIM with two carrier beams and a zero order beam.
    """

    def __init__(self):
        self._phaseStep = 5
        # Phase modulation on the beams (beam 0 is the zero order)
        self._phases = np.array([[0, 0, 0],
                       [0, 2 * np.pi / 5, -2 * np.pi / 5],
                       [0, 4 * np.pi / 5, -4 * np.pi / 5],
                       [0, 6 * np.pi / 5, -6 * np.pi / 5],
                       [0, 8 * np.pi / 5, -8 * np.pi / 5]])
        self._angleStep = 3
        # Angle of the beams in the pupil - set zero order to same angle as +/- orders.
        self._angles = np.array([[0, 0, 0],
                                 [2 * np.pi / 3, 2 * np.pi / 3, 2 * np.pi / 3],
                                 [4 * np.pi / 3, 4 * np.pi / 3, 4 * np.pi / 3]])
        # Angle of the beams relative to the illumination cone angle - zero order should have 0 here.
        self._alpha = np.array([[0, -1, 1],
                                [0, -1, 1],
                                [0, -1, 1]])
        # Amplitude of the three beams, same here for all astep and pstep.
        self._ampl = np.array([1, 1, 1])
        self._n_beams = 3
        self._nbands = 6
        super().__init__()

    def jones_vectors(self, astep):
        self.theta = np.arcsin(self.ill_NA / self.n)
        self.S = self.xp.zeros((self._n_beams, 3), dtype=self.xp.complex64)
        for i in range(self._n_beams):
            phi_S = self._angles[astep, i]
            f_p = self.xp.array(self.polarised_field(phi_S))
            self.S[i, :] = self._ampl[i] * self.xp.transpose(self.rotation(phi_S, self.theta * self._alpha[astep, i]) @ f_p)

    def _ill_obj_vec(self, x, y, z, pstep, astep):
        """Vectorised illumination intensity applied on the object"""
        p = self._phases[pstep, :]
        E = self.xp.zeros((self.npoints, self._n_beams, 3), dtype=self.xp.complex64)  # exponential terms of field
        xyz = self.xp.transpose(self.xp.stack([x, y, z]))
        for i in range(self._n_beams):
            phi_E = self._angles[astep, i]
            e = self.xp.exp(-1j * (xyz @
                                   self.rotation(phi_E, self.theta * self._alpha[astep, i]) @
                                   self.xp.array([0, 0, self.k0_ill]) + p[i]))
            E[:, i, :] = self.xp.transpose(self.xp.array([e, ] * 3))
        F = self.xp.sum(self.S * E, axis=1, dtype=self.xp.complex64)  # field of illumination
        # to calculate intensity: ill = F @ F_conjugate
        ill = (F * self.xp.conjugate(F)).real.round(15)  # take real part and round to 15 decimals
        return ill



#!c:/Users/maan/Documents/HexSimProcessor/venv/Scripts/python.exe

from src.napari_generic_simulator._widget import SIMulator, PointCloud

if __name__ == '__main__':
    import napari

    viewer = napari.Viewer()

    point_cloud = PointCloud(viewer)
    simulator = SIMulator(viewer)
    w1 = viewer.window.add_dock_widget(point_cloud, name='Point cloud generator', add_vertical_stretch=True)
    w2 = viewer.window.add_dock_widget(simulator, name='SIM data generator', add_vertical_stretch=True)
    viewer.window._qt_window.tabifyDockWidget(w1, w2)

    napari.run()



"""
Widget of the napari plugin
"""
__author__ = "Meizhu Liang @Imperial College London"

from magicgui import magicgui
from enum import Enum
from .baseSIMulator import import_cp, import_torch, torch_GPU
from .Illumination import ConIll, HexIll, RaHexIll, ConIll3D
from qtpy.QtWidgets import QWidget, QVBoxLayout, QFileDialog
from napari.qt.threading import thread_worker
from magicgui.widgets import SpinBox, Container, ComboBox, FloatSpinBox, LineEdit, RadioButtons, PushButton
from napari.layers import Layer
import tifffile
import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured
from pypcd_imp import pypcd
import matplotlib.pyplot as plt


class Samples(Enum):
    SPHEROIDAL = 500
    FILAMENTS = 10


class PointCloud(QWidget):
    """A widget to go with the SIMulator widget that could generate, display, load and save point clouds."""

    def __init__(self, viewer: 'napari.viewer.Viewer'):
        self._viewer = viewer
        super().__init__()
        self.setup_ui()

    def setup_ui(self):
        """Sets up the layout and adds the widget to the ui"""
        self.cloud_widgets()
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.add_magic_function(self.c_w, layout)

    def add_magic_function(self, function, _layout):
        """Adds the widget to the viewer"""
        self._viewer.layers.events.inserted.connect(function.reset_choices)
        self._viewer.layers.events.removed.connect(function.reset_choices)
        _layout.addWidget(function.native)

    def point_cloud_fil(self, nSamples, L, dL):
        """Generates a point-cloud as the object in the imaging system."""

        # L: full length of each filament in um
        # dL: target length of each short bit to form a filament

        for i in range(nSamples):
            alpha = 2 * np.pi * np.random.rand()
            beta = np.pi / 4 * (1 - 2 * np.random.rand())

            # centres of filaments
            xS = -L / 2 + L * np.random.rand()
            yS = -L / 2 + L * np.random.rand()
            zS = 0

            # starting points of filaments
            x1 = xS + (L / 2 * np.cos(alpha) * np.cos(beta))
            y1 = yS + (L / 2 * np.sin(alpha) * np.cos(beta))
            z1 = zS + (L / 2 * np.sin(beta))

            # ending points of filaments
            x2 = xS - (L / 2 * np.cos(alpha) * np.cos(beta))
            y2 = yS - (L / 2 * np.sin(alpha) * np.cos(beta))
            z2 = zS - (L / 2 * np.sin(beta))

            px = 2 * np.pi * np.random.rand()
            fx = 2 * np.pi * 0.5 * np.random.rand()
            ax = L / 10 * np.random.rand()
            py = 2 * np.pi * np.random.rand()
            fy = 2 * np.pi * 1.0 * np.random.rand()
            ay = L / 10 * np.random.rand()
            pz = 2 * np.pi * np.random.rand()
            fz = 2 * np.pi * 1.0 * np.random.rand()
            az = L / 10 * np.random.rand()

            # adding some perturbation
            l = np.arange(0, 1, dL / L)
            x = x1 + (x2 - x1) * l + ax * np.cos(fx * l + px)
            y = y1 + (y2 - y1) * l + ay * np.cos(fy * l + py)
            z = z1 + (z2 - z1) * l + az * np.cos(fz * l + pz)

            # x = x1 + (x2 - x1) * l
            # y = y1 + (y2 - y1) * l
            # z = z1 + (z2 - z1) * l
            xd = x[1:] - x[:-1]
            yd = y[1:] - y[:-1]
            zd = z[1:] - z[:-1]

            # steps = np.sqrt(xd ** 2 + yd ** 2 + zd ** 2)
            # print(f'length = {np.sum(steps):.2f} um, average = {1000 * np.mean(steps) :.2f} nm, stdev = {1000 * np.std(steps) :.2f} nm')
            if i == 0:
                points = np.single(np.dstack([x, y, z]).squeeze())
            else:
                points = np.concatenate((points, np.single(np.dstack([x, y, z]).squeeze())))
        return points

    def point_cloud(self, npoints, rad, dep):
        """Generates a point - cloud as the object in the imaging system."""

        # rad: radius of sphere of points
        # Multiply the points several times to get the enough number
        pointsxn = (2 * np.random.rand(npoints * 3, 3) - 1) * [rad, rad, rad]

        pointsxnr = np.sum(pointsxn * pointsxn, axis=1)  # multiple times the points
        points_sphere = pointsxn[pointsxnr < (rad ** 2), :]  # simulate spheres from cubes
        points = np.single(points_sphere[(range(npoints)), :])
        points[:, 2] = dep / rad * points[:, 2]  # to make the point cloud for OTF a ellipsoid rather than a sphere
        return points

    def gen_pc(self):
        np.random.seed(self.random_seed.value)
        if self.w_samples.value == Samples.FILAMENTS:
            self.pc = self.point_cloud_fil(self.fil_n.value, self.fil_len.value, self.fil_step.value)
            re_dep = self.fil_len.value / 2
            name = 'filaments'
        elif self.w_samples.value == Samples.SPHEROIDAL:
            self.pc = self.point_cloud(self.sph_points.value, self.sph_rad.value, self.sph_dep.value)
            re_dep = self.sph_dep.value
            name = 'spheroidal'
        print('Point cloud generated')
        if hasattr(self, 'pc'):
            try:
                self._viewer.add_points(-self.pc[:, ::-1], size=0.05, name=f'{re_dep}μm_{name}')
            except Exception as e:
                print(e)

    def save_pc(self):
        if hasattr(self._viewer.layers.selection.active, 'data'):
            try:
                filename, _ = QFileDialog.getSaveFileName(self, 'Save a file', filter='Files (*.pcd)')
                print(filename)
                pc = pypcd.make_xyz_point_cloud(-self._viewer.layers.selection.active.data)
                pypcd.save_point_cloud(pc, filename)
                print('Point cloud saved')
            except Exception as e:
                print(e)

    def load_pc(self):
        try:
            filename, _ = QFileDialog.getOpenFileName(self, 'Pick a file', filter='Files (*.pcd)')
            pc = pypcd.point_cloud_from_path(filename)
            out_arr = structured_to_unstructured(pc.pc_data)
            print(out_arr.shape, out_arr.dtype, out_arr)
            self._viewer.add_points(out_arr, size=0.1, name=filename[0])
        except Exception as e:
            print(e)

    def cloud_widgets(self):
        """Creates a widget containing all small widgets"""
        self.random_seed = SpinBox(value=123, name='spin', label='random seed')
        self.w_samples = RadioButtons(value=Samples.SPHEROIDAL, choices=Samples)

        self.sph_points = SpinBox(value=500, step=50, label='spheroidal_points')
        self.sph_rad = SpinBox(value=5, label='spheroidal_radius (μm)')
        self.sph_dep = FloatSpinBox(value=2.5, step=0.5, max=self.sph_rad.value, label='spheroidal_depth (μm)')
        self.w_sph = Container(widgets=[self.sph_points, self.sph_dep, self.sph_rad])

        self.fil_n = SpinBox(value=20, max=100, label='filament_n')
        self.fil_len = SpinBox(value=5, label='filament_length (μm)')
        self.fil_step = FloatSpinBox(value=0.05, step=0.01, max=self.fil_len.value, label='filament_step (μm)')

        self.w_fil = Container(widgets=[self.fil_n, self.fil_len, self.fil_step])

        self.comprehensive_w = Container(widgets=[magicgui(self.save_pc, call_button='Save current layer as .pcd',
                                                           auto_call=False),
                                                  magicgui(self.load_pc, call_button='Load point cloud',
                                                           auto_call=False)],labels=None)
        self.c_w = Container(widgets=[Container(widgets=[self.random_seed]), self.w_samples, self.w_sph, self.w_fil,
                                      Container(widgets=[magicgui(self.gen_pc,
                                                                  call_button='Generate point cloud',
                                                                  auto_call=False)], labels=None),
                                      self.comprehensive_w])

class Sim_mode(Enum):
    HEXSIM = 0
    HEXSIM_RA = 1
    SIM_CONV = 2  # Conventional 2-beam SIM
    SIM_CONV_3D = 3


class Pol(Enum):
    azimuthal = 0  # in_plane
    radial = 1  # axial
    circular = 2
    horizontal = 3
    vertical = 4


class Psf_calc(Enum):
    SCALAR = 0
    VEC_FLEXI = 1
    VEC_RIGID = 2


class Accel(Enum):
    NUMPY = 0
    if import_torch:
        TORCH_CPU = 1
        if torch_GPU:
            TORCH_GPU = 2
    if import_cp:
        CUPY = 3



class SIMulator(QWidget):
    """
    A Napari plugin for the simulation of raw images produced while scanning an object (3D point cloud) through focus as
    the hexSIM illumination pattern is shifted through 7 positions laterally.The raw data is processed by a standard, a
    frame-by-frame and a batch reconstruction to produce the super-resolved output.
    https://doi.org/10.1098/rsta.2020.0162
    This currently supports hexagonal SIM (1 angle, 7 phases) with three beams and that at right-angles.
    """

    def __init__(self, viewer: 'napari.viewer.Viewer'):
        self._viewer = viewer
        super().__init__()
        self.parameters()
        self.setup_ui()
        self.start_simulator()

    def setup_ui(self):
        """Sets up the layout and adds the widget to the ui"""
        self.wrap_widgets()
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.add_magic_function(self.w, layout)

    def add_magic_function(self, function, _layout):
        """Adds the widget to the viewer"""
        self._viewer.layers.events.inserted.connect(function.reset_choices)
        self._viewer.layers.events.removed.connect(function.reset_choices)
        _layout.addWidget(function.native)

    def parameters(self):
        self.SIM_mode = ComboBox(value=Sim_mode.SIM_CONV, label='SIM_mode', choices=Sim_mode)
        self.Polarisation = ComboBox(value=Pol.radial, label='Polarisation', choices=Pol)
        self.Acceleration = ComboBox(value=list(Accel)[-1], label='Acceleration', choices=Accel)
        self.Psf = ComboBox(value=Psf_calc.SCALAR, label='Psf calculation', choices=Psf_calc)
        self.N = SpinBox(value=128, name='spin', label='N pixel', max=1100)
        self.pixel_size = FloatSpinBox(value=6.5, name='spin', label='pixel size(μm)', step=0.5)
        self.magnification = SpinBox(value=60, name='spin', label='magnification')
        self.ill_NA = FloatSpinBox(value=1, name='spin', label='NA  illumination', min=0.0, step=0.1)
        self.det_NA = FloatSpinBox(value=1.0, name='spin', label='NA  detection', min=0.0, step=0.1)
        self.n = FloatSpinBox(value=1.33, name='spin', label='n', min=0.00)
        self.ill_wavelength = SpinBox(value=540, label='λ  illumination(nm)', step=10)
        self.det_wavelength = SpinBox(value=540, label='λ  detection(nm)', step=10)

        self.tpoints = SpinBox(value=140,  label='tpoints', min=0, max=500, step=1)
        self.xdrift = FloatSpinBox(value=0.0, label='xdrift(nm)', min=0.0, max=1000.0, step=5)
        self.drift = FloatSpinBox(value=0.0, label='Brownian motion(nm)', min=0.0, max=1000.0, step=5)
        self.sph_abb = FloatSpinBox(value=0.0, name='spin', label='spherical(rad)', min=-10.0, max=10, step=0.5)
        self.zchoice = ComboBox(value='zdrift(nm)', choices=['zdrift(nm)', 'zstep(nm)'], label=None)
        self.zmove = FloatSpinBox(value=50, min=0.0, max=10000.0, step=5)

    def par_list(self):
        """return the current parameter list"""
        return [self.SIM_mode.value, self.Polarisation.value, self.Acceleration.value,
                self.Psf.value, self.N.value, self.pixel_size.value, self.magnification.value, self.ill_NA.value,
                self.det_NA.value, self.n.value, self.ill_wavelength.value, self.det_wavelength.value, self.zrange,
                self.tpoints.value, self.xdrift.value, self.zchoice.value, self.zmove.value, self.drift.value, self.sph_abb.value]

    def set_att(self):
        """Sets attributes in the simulation class. Executed frequently to update the parameters"""
        if self.SIM_mode.value == Sim_mode.HEXSIM:
            # self.sim = HexSim_simulator()
            self.sim = HexIll()
            nsteps = self.sim._phaseStep * self.sim._angleStep
        elif self.SIM_mode.value == Sim_mode.HEXSIM_RA:
            self.sim = RaHexIll()
            nsteps = self.sim._phaseStep * self.sim._angleStep
        elif self.SIM_mode.value == Sim_mode.SIM_CONV:
            # self.sim = ConSim_simulator()
            self.sim = ConIll()
            nsteps = self.sim._phaseStep * self.sim._angleStep
        elif self.SIM_mode.value == Sim_mode.SIM_CONV_3D:
            self.sim = ConIll3D()
            nsteps = self.sim._phaseStep * self.sim._angleStep

        if self.Polarisation.value == Pol.azimuthal:
            self.sim.pol = 'a'
        elif self.Polarisation.value == Pol.radial:
            self.sim.pol = 'r'
        elif self.Polarisation.value == Pol.circular:
            self.sim.pol = 'c'
        elif self.Polarisation.value == Pol.horizontal:
            self.sim.pol = 'h'
        elif self.Polarisation.value == Pol.vertical:
            self.sim.pol = 'v'

        if self.Acceleration.value == Accel.NUMPY:
            self.sim.acc = 0

        if hasattr(Accel, 'TORCH_CPU'):
            if self.Acceleration.value == Accel.TORCH_CPU:
                self.sim.acc = 1
        if hasattr(Accel, 'TORCH_GPU'):
            if self.Acceleration.value == Accel.TORCH_GPU:
                self.sim.acc = 2
        if hasattr(Accel, 'CUPY'):
            if self.Acceleration.value == Accel.CUPY:
                self.sim.acc = 3

        if self.Psf.value == Psf_calc.VEC_RIGID:
            self.sim.psf_calc = 'vector_rigid'
        elif self.Psf.value == Psf_calc.VEC_FLEXI:
                self.sim.psf_calc = 'vector_flexible'
        elif self.Psf.value == Psf_calc.SCALAR:
            self.sim.psf_calc = 'scalar'

        self.sim.points = self.points
        self.sim.npoints = self.npoints

        self.sim.N = (self.N.value // 2) * 2
        self.N.value = self.sim.N
        self.sim.pixel_size = self.pixel_size.value
        self.sim.magnification = self.magnification.value
        self.sim.ill_NA = self.ill_NA.value
        self.sim.det_NA = self.det_NA.value
        self.sim.n = self.n.value
        self.sim.ill_wavelength = self.ill_wavelength.value * 1e-3
        self.sim.det_wavelength = self.det_wavelength.value * 1e-3
        self.sim.zrangeN = self.zrange * 2
        self.sim.tpoints = int((self.tpoints.value // nsteps / 2) * nsteps * 2)
        self.tpoints.value = self.sim.tpoints
        self.sim.xdrift = self.xdrift.value
        if self.zchoice.value == 'zdrift(nm)':
            self.sim.zdrift = self.zmove.value
            self.sim.zstep = False
        else:
            self.sim.zstep = self.zmove.value
            self.sim.zdrift = False
        self.sim.drift = self.drift.value
        self.sim.sph_abb = self.sph_abb.value
        self.used_par_list = [self.SIM_mode.value, self.Polarisation.value, self.Acceleration.value,
                              self.Psf.value,
                              self.N.value, self.pixel_size.value, self.magnification.value, self.ill_NA.value,
                              self.det_NA.value, self.n.value, self.ill_wavelength.value, self.det_wavelength.value,
                              self.zrange, self.tpoints.value, self.xdrift.value, self.zchoice.value,self.zmove.value,
                              self.drift.value, self.sph_abb.value]

    def start_simulator(self):
        """Starts the raw images generators and create the frequency space"""
        if hasattr(self, 'sim'):
            self.stop_simulator()
            self.start_simulator()
        else:
            pass

    def stop_simulator(self):
        if hasattr(self, 'sim'):
            delattr(self, 'sim')

    def select_layer(self, layer: Layer):
        """
        Selects a layer used to simulate raw SIM stacks, it contains the raw point-cloud data.
        Layer : napari.layers.Image
        """
        if not isinstance(layer, Layer):
            return
        if hasattr(self, 'points'):
            delattr(self, 'points')
        self.points = -layer.data[:, ::-1]
        self.npoints = self.points.shape[0]
        self.zrange = np.ceil(self.points[0].max() - self.points[0].min() + 0.5)
        self.messageBox.value = f'Selected sample layer: {layer.name}'
        self.get_results()

    def get_results(self):
        self.set_att()
        if not hasattr(self, 'points'):
            self.messageBox.value = f'Please select a point-cloud layer'
        else:
            if self.zchoice.value == 'zdrift(nm)':
                self.zsc = self.zrange / self.tpoints.value + self.zmove.value * 0.001  # z scale
                self.ztr = -self.zmove.value * 0.001 * self.tpoints.value / 2 - self.zrange / 2  # z translate
            else:
                self.zsc = (self.zrange / (self.tpoints.value / self.sim._nsteps) + self.zmove.value * 0.001) / self.sim._nsteps
                self.ztr = -self.zmove.value * 0.001 * (self.tpoints.value / self.sim._nsteps) / 2 - self.zrange / 2  # z translate
            md = {'mode': str(self.SIM_mode.value),
                      'pol': str(self.Polarisation.value),
                      'acc': str(self.Acceleration.value), 'psf': str(self.Psf.value),
                      'N': self.N.value, 'pix size': self.pixel_size.value,
                      'mag': self.magnification.value, 'ill NA': self.ill_NA.value,
                      'det NA': self.det_NA.value, 'n': self.n.value,
                      'ill_wavelength': self.ill_wavelength.value,
                      'det_wavelength': self.det_wavelength.value,
                      'z range': self.zrange,
                      'tpoints': self.tpoints.value, 'xdrift': self.xdrift.value,
                      'zChoice': self.zchoice.value, 'zmove': self.zmove.value,
                      'sph_abb': self.sph_abb.value}  # metadata
            xysc = self.pixel_size.value / self.magnification.value  # x or y scale
            def show_img(data):
                self._viewer.add_image(data, name='raw image stack',
                                       scale=(self.zsc, xysc, xysc),
                                       translate=(self.ztr,
                                                  -self.pixel_size.value / self.magnification.value * self.N.value / 2,
                                                  -self.pixel_size.value / self.magnification.value * self.N.value / 2),
                                       metadata=md)
                zr = (self.ztr, -self.ztr, -2 * self.ztr / self.tpoints.value - 1e-9)
                self._viewer.dims.range = (zr,  (
                    -xysc * self.N.value / 2, xysc * self.N.value / 2, xysc),
                                           (-xysc * self.N.value / 2, xysc * self.N.value / 2, xysc))
                self._viewer.dims.current_step = (data.shape[0] // 2, data.shape[1] // 2, data.shape[2] // 2)
                self._viewer.dims.axis_labels = ('z', 'x', 'y')
                delattr(self, 'points')

            @thread_worker(connect={"returned": show_img})
            def _get_results():
                t = self.sim.raw_image_stack_brownian()
                try:
                    while True:
                        self.messageBox.value = next(t)
                except Exception as e:
                    print(e)
                return self.sim.img

            _get_results()

    def wrap_widgets(self):
        # 'save and print' widgets
        save_tif_with_tags = PushButton(text='save_tif_with_tags')

        @save_tif_with_tags.clicked.connect
        def on_save_tif_with_tags_click():
            """Saves the selected image layer as a tif file with tags"""
            if hasattr(self._viewer.layers.selection.active, 'data'):
                try:
                    filename, _ = QFileDialog.getSaveFileName(self, "Save a file", filter='Images (*.tif)')
                    tifffile.imwrite(filename, self._viewer.layers.selection.active.data,
                                     description=str(self._viewer.layers.selection.active.metadata))
                except Exception as e:
                    print(str(e))

        show_psf = PushButton(text='show_psf')

        @show_psf.clicked.connect
        def on_show_psf_click():
            if hasattr(self, 'sim'):
                if hasattr(self.sim, 'psf'):
                    if self.used_par_list != self.par_list():
                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
                    else:
                        try:
                            if self.Psf.value == Psf_calc.VEC_RIGID:
                                self._viewer.add_image(self.sim.psf,
                                                   scale=(self.sim.dzn,
                                                          self.sim.dxn,
                                                          self.sim.dxn),
                                                   translate=(
                                                       -self.sim.dzn * self.sim.Nzn / 2,
                                                       -self.sim.dxn * self.sim.Nn / 2,
                                                       -self.sim.dxn * self.sim.Nn / 2),
                                                   name='3d-PSF', interpolation2d='spline36', colormap='viridis')
                                plt.figure()
                                plt.plot(np.sum(self.sim.psf_x, axis=(1, 2)))
                                plt.plot(np.sum(self.sim.psf_y, axis=(1, 2)))
                                plt.plot(np.sum(self.sim.psf_z, axis=(1, 2)))
                                plt.title(f'max(x) = {np.sum(self.sim.psf_x, axis=(1, 2)).max():.4f}, max(z) = {np.sum(self.sim.psf_z, axis=(1, 2)).max():.4f}')
                                plt.xlabel('z')
                                plt.ylabel('power sum (a.u.)')
                                plt.show()
                            else:
                                self._viewer.add_image(self.sim.psf,
                                                       scale=(self.sim.dzn,
                                                              self.sim.dxn,
                                                              self.sim.dxn),
                                                       translate=(
                                                           -self.sim.dzn * self.sim.Nzn / 2,
                                                           -self.sim.dxn * self.sim.Nn / 2,
                                                           -self.sim.dxn * self.sim.Nn / 2),
                                                       name='3d-PSF', interpolation2d='spline36', colormap='viridis')
                                plt.figure()
                                plt.plot(np.sum(self.sim.psf, axis=(1, 2)))
                                plt.title(f'max = {np.sum(self.sim.psf, axis=(1, 2)).max():.4f}')
                                plt.xlabel('z')
                                plt.ylabel('power sum (a.u.)')
                                plt.show()
                        except Exception as e:
                            print(e)

        print_tif = PushButton(text='print_tags')

        @print_tif.clicked.connect
        def on_print_tif_click():
            """Prints tags of the selected tif image"""
            try:
                if self._viewer.layers.selection.active.source.path is None:
                    print(self._viewer.layers.selection.active.metadata)
                else:
                    frames = tifffile.TiffFile(self._viewer.layers.selection.active.source.path)
                    page = frames.pages[0]
                    # Print file description
                    print(f'==={self._viewer.layers.selection.active.name}.tif===\n' + page.tags["ImageDescription"].value)
                self.messageBox.value = 'Parameters printed'
            except Exception as e:
                print(str(e))

        show_otf = PushButton(text='show_otf')
        @show_otf.clicked.connect
        def on_show_otf_click():
            if hasattr(self, 'sim'):
                if hasattr(self.sim, 'aotf'):
                    if self.used_par_list != self.par_list():
                        self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
                    else:
                        try:
                            self._viewer.add_image(self.sim.aotf,
                                                   scale=(self.sim.dzn,
                                                          self.sim.dxn,
                                                          self.sim.dxn),
                                                   translate=(
                                                       -self.sim.dzn * self.sim.Nzn / 2,
                                                       -self.sim.dxn * self.sim.Nn / 2,
                                                       -self.sim.dxn * self.sim.Nn / 2),
                                                   name='3d-OTF', colormap='viridis')
                        except Exception as e:
                            print(str(e))

        show_illumination = PushButton(text='illumination')

        @show_illumination.clicked.connect
        def on_show_illumination_click():
            if hasattr(self, 'sim'):
                if self.used_par_list != self.par_list():
                    self.messageBox.value = 'Parameters changed! Calculate the raw-image stack first!'
                else:
                    try:
                        self._viewer.add_image(self.sim.illumination_stack(),
                                               scale=(self.zsc,
                                                      self.pixel_size.value / self.magnification.value,
                                                      self.pixel_size.value / self.magnification.value),
                                               translate=(self.ztr,
                                                          -self.pixel_size.value / self.magnification.value * self.N.value / 2,
                                                          -self.pixel_size.value / self.magnification.value * self.N.value / 2),
                                               interpolation2d='spline36',
                                               name='illumination')
                        self._viewer.dims.current_step = (self.sim.illumination_stack().shape[0] // 2,
                                                          self.sim.illumination_stack().shape[1] // 2,
                                                          self.sim.illumination_stack().shape[2] // 2)
                    except Exception as e:
                        print(str(e))

        w_save_and_print = Container(widgets=[save_tif_with_tags, print_tif])
        w_show = Container(widgets=[show_psf, show_otf, show_illumination])
        self.messageBox = LineEdit(value='Messages')
        self.w = Container(widgets=[Container(widgets=[
            Container(widgets=[self.SIM_mode, self.Polarisation, self.Acceleration,
                               self.Psf, self.N,
                               self.ill_NA, self.det_NA, self.n, self.ill_wavelength, self.det_wavelength,
                               self.pixel_size,
                               self.magnification, self.tpoints,
                               self.xdrift, self.drift, self.sph_abb])]),
            Container(widgets=[self.zchoice, self.zmove], labels=None),
            magicgui(self.select_layer, call_button='Calculate results'),
            w_save_and_print, w_show, self.messageBox], labels=None)
